/*
 *  CallLog Jtapi Application
 *  Version 1.0     7/97
 *
 *  This sample application, uses Jtapi along with some of the Lucent's
 *  extensions to the API to access functionality specific to Lucent's
 *  DEFINITY G3 PBX.
 *  This application has three purposes:
 *
 *  1) Observe a terminal to log all incoming and outgoing calls
 *     to/from the specified device.  It also displays UserToUser
 *     information associated with the call.
 *  2) The ability to make calls and send UserToUser information along
 *     with the call
 *  3) The ability to send DTMF through an active call
 *
 *  Options:
 *
 *  Show Event Tracing - when this option is enabled, all events received
 *                       by the CallObserver on the terminal will be printed
 *                       in the tracing area.
 *  To execute, type:
 *
 *  java CallLog inputfile
 *
 *  where inputfile is an optional parameter that indicates a filename
 *  containing login information and phone directory.  The format of the data in
 *  this file should be as follows:
 *
 *  [Login]
 *  tlink=lucent#cstaserv#csta#myserver
 *  hostname=myserver.lucent.com
 *  login=user1
 *  password=user1
 *  device=3331
 *  [End]
 *
 *  [Directory]
 *  Audix=3333
 *  Home=787-222-2222
 *  user1=3331
 *  user2=3332
 *  [End]
 *
 */

import java.awt.*;
import java.util.*;
import javax.telephony.*;
import javax.telephony.media.*;
import javax.telephony.events.*;
import javax.telephony.callcontrol.*;
import javax.telephony.callcontrol.events.*;
import com.lucent.jtapi.tsapi.*;


/*
 * Class: CallLog
 *
 * Main class of the CallLog application.  It sets-up the user interface and defines
 * the  login, logout, makeCall and sendDTMF methods.
 *
 */
 public class CallLog extends Frame {
	void CallLog_GotFocus(Event event) {
		toFront();
	}

	void Login_Action(Event event) {
		//{{CONNECTION
		// verify if user wants to login or logout by checking menu item label
		if ( (loginMenuItem.getLabel()).equals("Login...") ) {
		    (new LoginDialog(this, "Tserver Login", true, trace, data)).show();
		} else {
		    logout();
		}
		//}}
	}

	void UUI_Action(Event event) {
		//{{CONNECTION
		// Action from Open... Show the OpenFileDialog
		(new UUIDialog(this, "Make Call with UUI", true, trace, phoneDirectory)).show();
		//}}
	}

	void DTMF_Action(Event event) {
		//{{CONNECTION
		(new DTMFDialog(this, "Send DTMF", true, trace)).show();
		//}}
	}

	void Disconnect_Action(Event event) {
		//{{CONNECTION
		disconnect();
		//}}
	}

	void Exit_Action(Event event) {
		//{{CONNECTION
		logout();
        hide();                 // hide the Frame
        dispose();              // free the system resources
        System.exit(0);         // close the application
        //}}
	}

	public CallLog() {

		//{{INIT_CONTROLS
		setLayout(null);
		addNotify();
		resize(insets().left + insets().right + 397,insets().top + insets().bottom + 185);
		setFont(new Font("Helvetica", Font.PLAIN, 11));
		setBackground(new Color(12632256));
		panel1 = new java.awt.Panel();
		panel1.setLayout(null);
		panel1.reshape(insets().left + 0,insets().top + 0,400,169);
		add(panel1);
		trace = new MyTextArea();
		trace.reshape(0,20,399,169);
		panel1.add(trace);
		setTitle("Jtapi Call Log");
		//}}

		//{{INIT_MENUS
		mainMenuBar = new java.awt.MenuBar();

		fileMenu = new java.awt.Menu("File");
		loginMenuItem = new java.awt.MenuItem("Login...");
		fileMenu.add(loginMenuItem);
		fileMenu.addSeparator();
		exitMenuItem = new java.awt.MenuItem("Exit");
		fileMenu.add(exitMenuItem);
		mainMenuBar.add(fileMenu);

		functionsMenu = new java.awt.Menu("Functions");
		uuiMenuItem = new java.awt.MenuItem("Make Call with UUI...");
		functionsMenu.add(uuiMenuItem);
		dtmfMenuItem = new java.awt.MenuItem("Send DTMF...");
		functionsMenu.add(dtmfMenuItem);
		disconnectMenuItem = new java.awt.MenuItem("Disconnect Active Call");
		functionsMenu.add(disconnectMenuItem);
		mainMenuBar.add(functionsMenu);

        optionsMenu = new java.awt.Menu("Options");
		tracingMenuItem = new java.awt.CheckboxMenuItem("Show event tracing");
		tracingMenuItem.setState(false);
		optionsMenu.add(tracingMenuItem);
		mainMenuBar.add(optionsMenu);

		functionsMenu.disable();
		setMenuBar(mainMenuBar);
		if ( !initJtapi() ) {
		    trace.appendText( "Error: JtapiPeer could not be created.  Verify your Jtapi client install.\n\n");
	        loginMenuItem.disable();
	    }
		//}}
	}

	public CallLog(String title) {
	    this();
	    setTitle(title);
	}

    public synchronized void show() {
    	move(50, 50);
    	super.show();
    }

	public boolean handleEvent(Event event) {
    	if (event.id == Event.WINDOW_DESTROY) {
            Exit_Action( event );
            return true;
    	}
		if (event.target == this && event.id == Event.GOT_FOCUS) {
			CallLog_GotFocus(event);
			return true;
		}
		return super.handleEvent(event);
	}

	public boolean action(Event event, Object arg) {
		if (event.target instanceof MenuItem) {
			if (event.target == loginMenuItem) {
				Login_Action(event);
				return true;
			} else if (event.target == uuiMenuItem) {
				UUI_Action(event);
				return true;
			} else if (event.target == dtmfMenuItem) {
				DTMF_Action(event);
				return true;
			} else if (event.target == disconnectMenuItem) {
				Disconnect_Action(event);
				return true;
			} else if (event.target == tracingMenuItem) {
				if ( tracingMenuItem.getState() ) {
				    trace.enableTracing();
				} else {
				    trace.disableTracing();
				}
				return true;
			} else if (event.target == exitMenuItem) {
					Exit_Action(event);
					return true;
			}
		}
		return super.action(event, arg);
	}

   /*
    * initJtapi()
    * This method creates JtapiPeer object
    *
    */
    public boolean initJtapi() {


        // create JtapiPeer object
        try {
		    // get Lucent's implementation of the JtapiPeer object
		    jtapiPeer = JtapiPeerFactory.getJtapiPeer( null );
        }
	    catch (JtapiPeerUnavailableException e) {
	        trace.appendText( "JtapiPeerFactory.getJtapiPeer: caught JtapiPeerUnavailableException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
	        return false;
	    }
	    return true;
    }

   /*
    * getServices()
    * This method uses the getServices method in the JtapiPeer interface to get
    * a list of the advertised tlinks.
    */
    public String[] getServices() {
	    String[]   services = null;

		// get tlinks
        try {
			services = jtapiPeer.getServices();
		}
		catch (PlatformException e) {
	        trace.appendText( "JtapiPeer.getServices(): caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
	        return null;
	    }
	    return services;
	}

   /*
    * login()
    * This method performs the following:
    *
    * 1) create Provider object
    * 2) verify if the terminal specified exists in the user's domain
    * 3) create Terminal object
    * 4) create Address object
    * 5) create CallObserver callback
    * 6) add Observer to Terminal using the addCallObserver method
    *
    */
    public void login(Hashtable args) {
        String          serviceID;
        String          login;
        String          password;
        String          providerString;
        String          callingExt;
        boolean         found = false;
        Terminal[]      observableTerminals = null;

        serviceID = (String) args.get( "serviceID" );
		login = (String) args.get( "login" );
		password = (String) args.get( "password" );
        callingExt = (String) args.get( "callingExt" );

		provider = null;
        providerString = serviceID + ";login=" + login + ";passwd=" + password;


		// 1) create Provider object
        try {
			if ( jtapiPeer != null ) {
			    provider = jtapiPeer.getProvider( providerString );
            } else {
                trace.appendText( "Error: JtapiPeer could not be created\n\n" );
                return;
            }

			if (provider != null) {
			    trace.appendText( "Provider created successfully\n\n" );
                trace.appendText( "Verifying that the provider is in service" );
    			while ( provider.getState() != Provider.IN_SERVICE )
    			{
    			    Thread.sleep( 3000 );
    			    trace.appendText( "." );
    			}
    			trace.appendText( "\nProvider is in service\n\n" );
    			provider.addObserver( new MyProviderObserver( this ) );
    			uuiMenuItem.enable();
		        loginMenuItem.setLabel( "Logout..." );
    	    } else {
    	        trace.appendText( "Error: Provider could not be created\n\n" );
    	        return;
    	    }
		}
		catch (PlatformException e) {
	        trace.appendText( "JtapiPeer.createProvider(): caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
	        return;
	    }
        catch (Exception e) {
	        trace.appendText( "JtapiPeer.createProvider(): caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
	        return;
	    }

        // 2) verify if the terminal specified exists in this user's domain
       /*
        * This step is not necessary because our user interface is asking for
        * specific terminal/address.  Otherwise, we could use the getTerminals() or
        * getAddresses() methods in the Provider interface to check what devices
        * the user has control over.
        */
       /*
        try {
            trace.appendText( "Please wait while the Provider obtains the terminals in its domain...\n\n" );
    		observableTerminals = provider.getTerminals();
	    }
	    catch (ResourceUnavailableException e) {
	        trace.appendText( "Provider.getTerminals(): caught ResourceUnavailableException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    if ( observableTerminals != null ) {
	        for ( int i=0; i < observableTerminals.length; i++ ) {
	            if ( callingExt.equals ( observableTerminals[i].getName() ) ) {
	                found = true;
	                break;
	            }
	        }
	        if ( !found ) {
	            trace.appendText( "Error: Terminal " + callingExt + " doesn't exist in the Security Database or\n" +
	                              "the user doesn't have permissions to observe it.\n\n");
                return;
            }
        } else {
            trace.appendText( "Error: There are no terminals in this Provider's domain.\n" +
	                          "Check the Tserver's administration.\n\n" );
	        return;
        }
        */

        // 3) create Terminal object
        try {
    	    myTerminal = provider.getTerminal(callingExt);
        }
	    catch (InvalidArgumentException e) {
	        trace.appendText("Provider.getAddress(): caught InvalidArgumentException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (PlatformException e) {
	        trace.appendText( "Provider.getAddress(): caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }

	    if ( myTerminal != null ) {
		    trace.appendText( "Terminal " + callingExt + " created successfully\n\n" );

        } else {
            trace.appendText( "Terminal " + callingExt + " could not be created\n\n" );
            return;
        }

        // 4) create Address object
        try {
            myAddress = provider.getAddress( callingExt );
        }
	    catch (InvalidArgumentException e) {
	        trace.appendText( "Provider.getAddress(): caught InvalidArgumentException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (PlatformException e) {
	        trace.appendText( "Provider.getAddress(): caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    if ( myAddress != null ) {
		    trace.appendText( "Address " + callingExt + " created successfully\n\n" );
        } else {
            trace.appendText( "Address " + callingExt + " could not be created\n\n" );
            return;
        }


        // 5) create CallObserver callback
        myCallObserver = new MyCallObserver( this );
        trace.appendText( false, "Adding callObserver to terminal\n\n" );
        trace.appendText( "Logged in successfully...\n\n" );
        functionsMenu.enable();

        // 6) Add CallObserver to Terminal
        try {
            myTerminal.addCallObserver(myCallObserver);
        }
        catch (ResourceUnavailableException e) {
	        trace.appendText("Terminal.addCallObserver(): caught ResourceUnavailableException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (PlatformException e) {
	        trace.appendText("Terminal.addCallObserver(): caught PlatformException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
    }

   /*
    * makeCall() - makes a call with UserToUserInfo to specified device
    * This method creates  the Call object and uses the connect() method
    * from the LucentCall interface to place the call to the destination
    * number with any UUI specified.
    */
    public void makeCall(Hashtable args) {
        String                  calledString;
        String                  calledNumber;
        String                  uui;
        Call                    call = null;
        Connection[]            connections = null;
        TerminalConnection[]    terminalConnections = null;
        UserToUserInfo          lucentUUI = null;

		calledString = (String) args.get("calledString");
        uui = (String) args.get("uui");

        if ( ( phoneDirectory != null ) &&
             ( phoneDirectory.containsKey ( calledString ) ) ) {
                calledNumber = (String) phoneDirectory.get ( calledString );
        } else {
            calledNumber = calledString;
        }

        // create UserToUserInfo object
        if ( uui != null ) {
            lucentUUI = new UserToUserInfo( uui );
        }

        try {
            // create Call object
            call = provider.createCall();
        }
	    catch (InvalidStateException e) {
	        trace.appendText("Provider.createCall(): caught InvalidStateException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (MethodNotSupportedException e) {
	        trace.appendText("Provider.createCall(): caught MethodNotSupportedException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (ResourceUnavailableException e) {
	        trace.appendText("Provider.createCall(): caught ResourceUnavailableException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (PlatformException e) {
	        trace.appendText("Provider.createCall(): caught PlatformException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
	    catch (PrivilegeViolationException e) {
	        trace.appendText("Provider.createCall(): caught PrivilegeViolationException\n");
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }

        if (call != null) {
            trace.appendText(false, "Call created successfully\n\n");
        } else {
            trace.appendText("Error: Call could not be created\n\n");
            return;
        }

        try {
    	    // makecall by using the connect method in the LucentCall interface
            connections = ( (LucentCall) call).connect(myTerminal, myAddress, calledNumber, false, lucentUUI);
    	}
        catch ( InvalidStateException e ) {
            trace.appendText ( "Call.connect: caught InvalidStateException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( MethodNotSupportedException e ) {
            trace.appendText ( "Call.connect: caught MethodNotSupportedException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( InvalidArgumentException e ) {
            trace.appendText ( "Call.connect: caught InvalidArgumentException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( ResourceUnavailableException e ) {
            trace.appendText ( "Call.connect: caught ResourceUnavailableException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( PlatformException e ) {
            trace.appendText ( "Call.connect: caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( PrivilegeViolationException e ) {
            trace.appendText ( "Call.connect: caught PrivilegeViolationException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        catch ( InvalidPartyException e ) {
            trace.appendText ( "Call.connect: caught InvalidPartyException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
    }

   /*
    * sendDTMF() - sends DTMF on active call (in the TALKING state)
    * This method uses the getTerminalConnections method of the Terminal interface
    * to get a handle on the connection that is in the TALKING state.  When a connection
    * is found, then uses the generateDtmf method of the MediaTerminalConnection
    * interface to send DTMF.
    */
    public void sendDTMF(Hashtable args) {
        String                              dtmf;
		CallControlTerminalConnection[]     terminalConnections = null;
		boolean                             found = false;

		dtmf = (String) args.get("dtmf");

        try {
            terminalConnections = (CallControlTerminalConnection[]) myTerminal.getTerminalConnections();
        }
        catch ( PlatformException e ) {
            trace.appendText ( "Terminal.getTerminalConnections: caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }
        if ( terminalConnections != null ) {
            for (int i=0; i < terminalConnections.length; i++ ) {
                if ( terminalConnections[i].getCallControlState() ==
                     CallControlTerminalConnection.TALKING ) {
                     found = true;
                    try {
                        ((MediaTerminalConnection) terminalConnections[i]).generateDtmf( dtmf );
                    }
                    catch ( InvalidStateException e ) {
                        trace.appendText ( "LucentTerminalConnection.generateDtmf: caught InvalidStateException\n" );
                        trace.appendText( "Message: " + e.getMessage() + "\n\n" );
            	        e.printStackTrace();
                    }
                    catch ( MethodNotSupportedException e ) {
                        trace.appendText ( "LucentTerminalConnection.generateDtmf: caught MethodNotSupportedException\n" );
                        trace.appendText( "Message: " + e.getMessage() + "\n\n" );
            	        e.printStackTrace();
                    }
                    catch ( ResourceUnavailableException e ) {
                        trace.appendText ( "LucentTerminalConnection.generateDtmf: caught ResourceUnavailableException\n" );
                        trace.appendText( "Message: " + e.getMessage() + "\n\n" );
            	        e.printStackTrace();
                    }
                    catch ( PlatformException e ) {
                        trace.appendText ( "LucentTerminalConnection.generateDtmf: caught PlatformException\n" );
                        trace.appendText( "Message: " + e.getMessage() + "\n\n" );
            	        e.printStackTrace();
                    }
                    finally {
                        break;
                    }
                }
            }
            if ( !found ) {
                trace.appendText ( "There are no Connections in the Talking state\n\n" );
            }
        }
    }

   /*
    * disconnect() - Disconnects active call (in the TALKING state)
    * This method uses the getTerminalConnections method of the Terminal interface.
    * If there is only one terminalConnection, terminate that one.  Otherwise,
    * get a handle on the connection that is in the TALKING state.  When a connection
    * is found, then calls the disconnect(connection) method to end the call.
    */
    public void disconnect() {
		TerminalConnection[]     terminalConnections = null;
		boolean                  found = false;
		Connection               connection;

        try {
            terminalConnections = myTerminal.getTerminalConnections();
        }
        catch ( PlatformException e ) {
            trace.appendText ( "Terminal.getTerminalConnections: caught PlatformException\n" );
            trace.appendText( "Message: " + e.getMessage() + "\n\n" );
	        e.printStackTrace();
        }

        if ( terminalConnections != null ) {

            if ( terminalConnections.length == 1 ) {

                connection = terminalConnections[0].getConnection();
                disconnect ( connection );

            } else {
                for ( int i=0; i < terminalConnections.length; i++ ) {

                    if ( ( (CallControlTerminalConnection) terminalConnections[i]).getCallControlState() ==
                         CallControlTerminalConnection.TALKING ) {

                        found = true;
                        connection = terminalConnections[i].getConnection();

                        disconnect ( connection );
                    }
                }
                if ( !found ) {
                    trace.appendText ( "There are no Connections in the Talking state\n\n" );
                }
            }
        }
    }

    /*
    * disconnect(Connection) - Disconnects connection passed in argument
    * This method uses the disconnect method of the Connection interface to end the call.
    */
    public void disconnect( Connection connection ) {


        if ( connection != null ) {

            try {
                connection.disconnect();
            }
            catch ( PrivilegeViolationException e ) {
                trace.appendText ( "Connection.disconnect: caught PrivilegeViolationException\n" );
                trace.appendText( "Message: " + e.getMessage() + "\n\n" );
    	        e.printStackTrace();
            }
            catch ( InvalidStateException e ) {
                trace.appendText ( "Connection.disconnect: caught InvalidStateException\n" );
                trace.appendText( "Message: " + e.getMessage() + "\n\n" );
    	        e.printStackTrace();
            }
            catch ( MethodNotSupportedException e ) {
                trace.appendText ( "Connection.disconnect: caught MethodNotSupportedException\n" );
                trace.appendText( "Message: " + e.getMessage() + "\n\n" );
    	        e.printStackTrace();
            }
            catch ( ResourceUnavailableException e ) {
                trace.appendText ( "Connection.disconnect: caught ResourceUnavailableException\n" );
                trace.appendText( "Message: " + e.getMessage() + "\n\n" );
    	        e.printStackTrace();
            }
            catch ( PlatformException e ) {
                trace.appendText ( "Connection.disconnect: caught PlatformException\n" );
                trace.appendText( "Message: " + e.getMessage() + "\n\n" );
    	        e.printStackTrace();
            }
        }
    }


   /*
    * logout() - execute a Provider shutdown, and reset menus
    * This method first verifies that the Provider is in service, it then removes the
    * callObserver that we have on the terminal, and then uses the shutdown method
    * on the Provider to terminate session.
    */
    public void logout() {

        if (provider != null ) {

		    if ( provider.getState() == Provider.IN_SERVICE ) {

                // remove callObserver on terminal (nice but not necessary)
                if ( myTerminal != null ) {
                    try {
                        CallObserver observers[] = myTerminal.getCallObservers();

                        if ( observers != null ) {

                            for ( int i=0; i < observers.length; i++ ) {
                                myTerminal.removeCallObserver( observers[i] );
                            }
                        }
                    }
                    catch ( PlatformException e ) {
                        trace.appendText ( "Terminal.removeCallObserver(): caught PlatformException\n" );
                        trace.appendText( "Message: " + e.getMessage() + "\n\n" );
            	        e.printStackTrace();
                    }
                }

                // shutdown Provider
                trace.appendText("Shutting down provider...\n\n");
        		try {
        		    provider.shutdown();
        		}
                catch ( PlatformException e ) {
                    trace.appendText ( "ProviderFrame.shutdown(): caught PlatformException\n" );
                    trace.appendText( "Message: " + e.getMessage() + "\n\n" );
        	        e.printStackTrace();
                }
                finally {
                    provider = null;
                }
            }
        }

        // reset menus
        handleProviderShutdown( null );
	}

	/*
    * handleConnActive()
    * When there is a connection active, enable the "Send DTMF" and the "Disconnect" menu
    * items; disable the "Make Call with UUI" menu item
    */
    public void handleConnActive(CallEv event) {

        dtmfMenuItem.enable();
        disconnectMenuItem.enable();
        uuiMenuItem.disable();
    }

    /*
    * handleConnDropped()
    * When a connection has gone away, disable the "Send DTMF" and the "Disconnect" menu
    * items; enable the "Make Call with UUI" menu item
    */
    public void handleConnDropped(CallEv event) {

        dtmfMenuItem.disable();;
        disconnectMenuItem.disable();
        uuiMenuItem.enable();
    }

   /*
    * handleOutgoingCall()
    * When a call has been placed from our observed terminal, record:
    *   - date and time call was placed
    *   - the called device information
    *   - UserToUserInfo sent, if any,  by using the getUserToUserInfo method in the
    *     LucentCallInfo interface.
    * Note that this method will be called for two specific events:
    *   - CallCtlConnAlertingEv - to track when connection is alerting at terminating
    *                             end of the call for internal calls
    *   - CallCtlConnNetworkReachedEv - to track when connection has reached the network
    *                             for external calls, in this case the UUI is not available
    */
    public void handleOutgoingCall( CallEv event ) {

        Terminal        callingTerminal;
		Address         callingAddress;
        Address         calledAddress;
        UserToUserInfo  lucentUUI = null;
        String          calledNumber = "<Unknown>";
        Address         address;
        Call            call;

        callingTerminal = ( (CallCtlCallEv) event).getCallingTerminal();

        // first verify that the callingTerminal is our terminal
        if ( myTerminal.equals ( callingTerminal ) ) {

            address = ((ConnEv) event).getConnection().getAddress();
            callingAddress = ( (CallCtlCallEv) event).getCallingAddress();
		    calledAddress = ( (CallCtlCallEv) event).getCalledAddress();

            if ( address != null && callingAddress != null ) {

                // verify if this CallCtlCallEv belongs to the terminating end of the call
                if ( !( address.equals ( callingAddress ) ) ) {

                    Date d = new Date();

                    // get called number information
                    if ( calledAddress != null ) {
                        calledNumber = calledAddress.getName();
                    }

                    // log call
                    trace.appendText( d.toLocaleString() + " Call made to: " + calledNumber );

            		call = event.getCall();
            		// verify if there is UUI along with this call
            		if ( call instanceof LucentCallInfo ) {

            		    if ( event.getID() == CallCtlConnNetworkReachedEv.ID ) {
            		        trace.appendText("  UUI: <unknown>\n\n");
            		    } else {
            		        lucentUUI = ( (LucentCallInfo) call ).getUserToUserInfo();
                		    if ( lucentUUI != null ) {
                		        String uui = lucentUUI.getString();
                		        trace.appendText("  UUI: " + uui + "\n\n");
                		    } else {
                		        trace.appendText("  UUI: <none>\n\n");
                		    }
                		}
            		}
        		}
        	}
        }
    }

   /*
    * handleIncomingCall()
    * When a call has been received by our observed terminal, record:
    *   - date and time call was received
    *   - the calling device information
    *   - UserToUserInfo received, if any,  by using the getUserToUserInfo method in the
    *     LucentCallInfo interface.
    * Note that this method will be called for two specific events:
    *   - CallControlTermConnRingingEv - to track when calls are alerting at our observed
    *                                    terminal
    */
    public void handleIncomingCall(CallEv event) {

        Address         callingAddress;
        UserToUserInfo  lucentUUI = null;
        String          uui = null;
        Date            d = new Date();

        callingAddress = ( (CallCtlCallEv) event).getCallingAddress();

        if ( event instanceof LucentCallInfo ) {
		    lucentUUI = ( (LucentCallInfo) event ).getUserToUserInfo();
		    if ( lucentUUI != null ) {
		        uui = "  UUI: " + lucentUUI.getString() + "\n\n";
		    } else {
		        uui = "  UUI: <none>\n\n";
		    }
		}

        if ( callingAddress != null ) {

            // verify that callingAddress is not our address, otherwise, it is us
            // initiating the call
            if ( !( myAddress.equals( callingAddress )) ) {

                trace.appendText(d.toLocaleString() + " Call received from: " +
                                                        callingAddress.getName() + uui);
            }
        } else {
            trace.appendText(d.toLocaleString() + " Call received from: <Unknown>" + uui);
        }
    }

   /*
    * handleProvOutOfService()
    * This method handles the case when our Provider went out of service for any
    * reason.  It loops until the Provider is back in service using the getState
    * method of the Provider interface.
    */
    public void handleProvOutOfService(ProvEv event) {

        Provider provider = event.getProvider();
        if ( provider != null ) {
            trace.appendText("Waiting for the provider to be back in service");
			while (provider.getState() != Provider.IN_SERVICE)
			{
			    try {
			        Thread.sleep(3000);
			    } catch (InterruptedException e) {
        	        trace.appendText("Thread.sleep(): caught InterruptedException\n");
                    trace.appendText( "Message: " + e.getMessage() + "\n\n" );
        	        e.printStackTrace();
                }
    	        trace.appendText(".");
			}
			trace.appendText("\nProvider is back in service\n\n");
        }
    }

   /*
    * handleProviderShutdown()
    * This method does the necessary clean-up for our application, when the session has been
    * terminated.
    */
    public void handleProviderShutdown(ProvEv event) {

        provider = null;
        loginMenuItem.setLabel("Login...");
        functionsMenu.disable();
    }


    public void paint( Graphics d ) {
    	Dimension size;
    	panel1.resize(size());
    	size = size();
    	trace.resize( size.width - 8, size.height - 46 );
    	return;
    }

   /*
    * readPreferences()
    * This method is called when an argument has been specified when the program
    * was called.  This argument is a filename with default values of login information.
    * It uses the Preferences class to invoke the getData method which returns a
    * Hashtable containing parameter-value pairs
    */
    public static void readPreferences(String  filename) {

        Preferences prefs = new Preferences( trace );
	    data = prefs.getData( filename, "[Login]" );
	    phoneDirectory = prefs.getData( filename, "[Directory]" );
	    prefs.trimObjects ( phoneDirectory, " -()." );
    }


	static public void main( String args[] ) {
	    (new CallLog()).show();
	    if ( args.length >= 1 ) {
	        readPreferences( args[0] );
	    }
	}

	//{{DECLARE_CONTROLS
	java.awt.Panel panel1;
	//}}

	//{{DECLARE_MENUS
	java.awt.MenuBar mainMenuBar;
	java.awt.Menu fileMenu;
	java.awt.MenuItem exitMenuItem;
	java.awt.Menu functionsMenu;
	java.awt.Menu optionsMenu;
	java.awt.Menu helpMenu;
	java.awt.MenuItem aboutMenuItem;
	//}}
	static MyTextArea   trace;
	Provider            provider;
	Terminal            myTerminal;
	Address             myAddress;
	MyCallObserver      myCallObserver;
	java.awt.MenuItem   uuiMenuItem;
	java.awt.MenuItem   dtmfMenuItem;
	java.awt.MenuItem   loginMenuItem;
	java.awt.MenuItem   disconnectMenuItem;
	CheckboxMenuItem    tracingMenuItem;
	static Hashtable    data;
	static Hashtable    phoneDirectory;
    JtapiPeer           jtapiPeer;
}


/*
 * Class: EventHandler
 * This class is the thread that handles the actions to be taken from events recevived
 * by our CallObserver and ou ProviderObserver.  We should not be calling this functions
 * directly from our callbacks.
 */
 class EventHandler extends Thread {

    public EventHandler ( Ev ev, Frame frame ) {
        super();
        if ( frame instanceof CallLog ) {
            callLogApp = ( CallLog) frame;
        }
        event = ev;
    }

    public void run() {

        if ( callLogApp != null ) {
            switch ( event.getID() )
            {
                case TermConnActiveEv.ID:
    	            callLogApp.handleConnActive( (CallEv) event );
    	            break;
    	        case TermConnDroppedEv.ID:
    	            callLogApp.handleConnDropped( (CallEv) event );
    	            break;
    	        case CallCtlConnAlertingEv.ID:
    	            callLogApp.handleOutgoingCall( (CallEv) event );
    	            break;
    	        case CallCtlConnNetworkReachedEv.ID:
    	            callLogApp.handleOutgoingCall( (CallEv) event );
    	            break;
    	        case CallCtlTermConnRingingEv.ID:
    	            callLogApp.handleConnActive( (CallEv) event );
    	            callLogApp.handleIncomingCall( (CallEv) event );
    	            break;
    	        case ProvOutOfServiceEv.ID:
    	            callLogApp.handleProvOutOfService( (ProvEv) event );
    	            break;
    	        case ProvShutdownEv.ID:
    	            callLogApp.handleProviderShutdown( (ProvEv) event );
    	            break;
    	    }
    	}
    }

    CallLog     callLogApp;
    Ev          event;
 }



/*
 * Class: MyCallObserver
 *
 * This class implements the CallObserver and CallControlCallObserver interfaces.
 * It reports all state changes on the Call object as events.  These events are
 * used to handle incoming and outgoing calls appropriately.
 *
 */
 class MyCallObserver implements CallObserver, CallControlCallObserver
 {
    public MyCallObserver() {
    }

    public MyCallObserver( Frame _frame ) {
        this();
        frame = _frame;
        if ( frame instanceof CallLog ) {
            msgTrace = ((CallLog) frame).trace;
        }
    }

    public void callChangedEvent( CallEv[] eventList )
    {
        Terminal        callingTerminal;
		Address         calledAddress;
        UserToUserInfo  lucentUUI = null;
        int             i = 0;

        msgTrace.appendText ( false, "\n" );
        for ( i = 0; i < eventList.length; i++ )
        {
	        if (eventList[i].isNewMetaEvent()) {
			    msgTrace.appendText( false,	"NEW META EVENT________________________________________\n");
			}

            msgTrace.appendText( false, "Received ");
            switch(eventList[i].getID())
            {
            case CallActiveEv.ID:
	            msgTrace.appendText( false, "Call Active Event\n" );
	            break;
            case CallInvalidEv.ID:
	            msgTrace.appendText( false, "Call Invalid Event\n" );
	            break;
            case CallObservationEndedEv.ID:
	            msgTrace.appendText( false, "Call Observation Ended Event\n" );
	            break;
            case ConnAlertingEv.ID:
	            msgTrace.appendText( false, "Connection Alerting Event\n" );
	            break;
	        case ConnInProgressEv.ID:
	            msgTrace.appendText( false, "Connection in Progress Event\n" );
	            break;
	        case ConnConnectedEv.ID:
	            msgTrace.appendText( false, "Connection Connected Event\n" );
	            break;
	        case ConnCreatedEv.ID:
	            msgTrace.appendText( false, "Connection Created Event\n" );
	            break;
	        case ConnDisconnectedEv.ID:
	            msgTrace.appendText( false, "Connection Disconnected Event\n" );
	            break;
	        case ConnFailedEv.ID:
        		msgTrace.appendText( false, "Connection Failed Event\n" );
	            break;
            case ConnUnknownEv.ID:
        		msgTrace.appendText( false, "Connection Unknown Event\n" );
	            break;
	        case TermConnActiveEv.ID: //spawn thread to handle active call
        		msgTrace.appendText( false, "TerminalConnection Active Event\n" );
	            (new EventHandler ( eventList[i], frame )).start();
	            break;
	        case TermConnCreatedEv.ID:
        		msgTrace.appendText( false, "TerminalConnection Created Event\n" );
	            break;
	        case TermConnDroppedEv.ID: //spawn thread to handle call going away
        		msgTrace.appendText( false, "TerminalConnection Dropped Event\n" );
	            (new EventHandler ( eventList[i], frame )).start();
	            break;
	        case TermConnPassiveEv.ID:
        		msgTrace.appendText( false, "TerminalConnection Passive Event\n" );
	            break;
	        case TermConnRingingEv.ID:
        		msgTrace.appendText( false, "TerminalConnection Ringing Event\n" );
	            break;
	        case TermConnUnknownEv.ID:
        		msgTrace.appendText( false, "TerminalConnection Unknown Event\n" );
	            break;
            case CallCtlConnAlertingEv.ID: //spawn thread to handle outgoing call
        		msgTrace.appendText( false, "CallCtlConnAlertingEv\n" );
	            (new EventHandler ( eventList[i], frame )).start();
	            break;
	        case CallCtlConnDialingEv.ID:
        		msgTrace.appendText( false, "CallCtlConnDialingEv\n" );
	            break;
	        case CallCtlConnDisconnectedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnDisconnectedEv\n" );
	            break;
	        case CallCtlConnEstablishedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnEstablishedEv\n" );
	            break;
	        case CallCtlConnFailedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnFailedEv\n" );
	            break;
	        case CallCtlConnInitiatedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnInitiatedEv\n" );
	            break;
	        case CallCtlConnNetworkAlertingEv.ID:
        		msgTrace.appendText( false, "CallCtlConnNetworkAlertingEv\n" );
	            break;
	        case CallCtlConnNetworkReachedEv.ID: //spawn thread to handle outgoing call
        		msgTrace.appendText( false, "CallCtlConnNetworkReachedEv\n" );
	            (new EventHandler ( eventList[i], frame )).start();
	            break;
	        case CallCtlConnOfferedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnOfferedEv\n" );
	            break;
	        case CallCtlConnQueuedEv.ID:
        		msgTrace.appendText( false, "CallCtlConnQueuedEv\n" );
	            break;
	        case CallCtlConnUnknownEv.ID:
        		msgTrace.appendText( false, "CallCtlConnUnknownEv\n" );
	            break;
	        case CallCtlTermConnBridgedEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnBridgedEv\n" );
	            break;
	        case CallCtlTermConnDroppedEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnDroppedEv\n" );
	            break;
	        case CallCtlTermConnHeldEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnHeldEv\n" );
	            break;
	        case CallCtlTermConnInUseEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnInUseEv\n" );
	            break;
	        case CallCtlTermConnRingingEv.ID: //spawn thread to handle incoming call
        		msgTrace.appendText( false, "CallControlTermConnRingingEv\n" );
	            (new EventHandler ( eventList[i], frame )).start();
        		break;
	        case CallCtlTermConnTalkingEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnTalkingEv\n" );
        		break;
	        case CallCtlTermConnUnknownEv.ID:
        		msgTrace.appendText( false, "CallCtlTermConnUnknownEv\n" );
	            break;
	        default:
	            msgTrace.appendText( false, "Unknown CallEv: " + eventList[i].getID() + "\n" );
	            break;
		    }
		}
   	}

   	Frame       frame;
   	MyTextArea  msgTrace;
 }

/*
 * Class: MyProviderObserver
 *
 * This class implements the ProviderObserver interface.
 * It reports all state changes on the Provider object as events.  These events are
 * used to know when the Provider is able to service us or not.
 *
 */
 class MyProviderObserver implements ProviderObserver
 {
    public MyProviderObserver() {
    }

    public MyProviderObserver(Frame _frame) {
        this();
        frame = _frame;
        if ( frame instanceof CallLog ) {
            msgTrace = ((CallLog) frame).trace;
        }
    }

    public void providerChangedEvent(ProvEv[] eventList)
    {
        int i = 0;

	    msgTrace.appendText( "\n" );
	    for ( i=0; i<eventList.length; i++ )
        {
	        if (eventList[i].isNewMetaEvent()) {
			    msgTrace.appendText( false,	"NEW META EVENT________________________________________\n");
			}

            msgTrace.appendText( "Received " );
	        switch(eventList[i].getID())
            {
            case ProvInServiceEv.ID:
	            msgTrace.appendText( false, "Provider In Service Event\n" );
	            break;
            case ProvOutOfServiceEv.ID:
	            msgTrace.appendText( "Provider Out Of Service Event\n" );
	            (new EventHandler ( eventList[i], frame )).start();
        		break;
            case ProvObservationEndedEv.ID:
	            msgTrace.appendText( false, "Provider Observation Ended Event\n" );
	            (new EventHandler ( eventList[i], frame )).start();
        		break;
            case ProvShutdownEv.ID:
	            msgTrace.appendText( "Provider Shutdown Event\nPlease re-login to the Tserver.\n" );
	            (new EventHandler ( eventList[i], frame )).start();
        		break;
	        default:
	            msgTrace.appendText( "Unknown: " + eventList[i].getID() + "\n" );
	            break;
		    }

		}
    }

    Frame               frame;
    MyTextArea          msgTrace;
}